package fmon.battle

import javafx.fxml.FXML
import javafx.fxml.FXMLLoader
import javafx.fxml.JavaFXBuilderFactory
import javafx.scene.{control => jfxsc}
import javafx.scene.image.ImageView

import scalafx.Includes._
import scalafx.animation._
import scalafx.geometry.{Point2D, Pos}
import scalafx.scene.{Group, Parent}
import scalafx.scene.control._
import scalafx.scene.layout._
import scalafx.scene.paint.Color
import scalafx.scene.shape.{Circle, Polygon}
import scalafx.scene.text.Font
import scalafx.scene.transform.Rotate
import scalafx.util.Duration

import fmon.battle.msg._
import fmon.stat._
import fmon.stat.Statistic._

class BattleUI extends SignalConsumer {
  @FXML var buttonPane: javafx.scene.layout.VBox = _
  @FXML var messages: jfxsc.TextArea = _
  @FXML var animationPane: javafx.scene.layout.Pane = _
  
  @FXML var backgroundImg: ImageView = _
  @FXML var playerImg: ImageView = _
  @FXML var enemyImg: ImageView = _
  
  @FXML var playerMonController: InfoWidget = _
  @FXML var enemyMonController: InfoWidget = _
  
  private var engine: BattleEngine = _
  
  def setEngine(engine: BattleEngine): Unit = {
    this.engine = engine
    updateUI()
  }
  
  def updateUI(): Unit = {
    val player = engine.player.lead
    val enemy = engine.enemy.lead
    val buttons = player.base.moves.map(move => loadMoveButton(move))
    buttonPane.children = buttons
    
    playerMonController.update(player)
    enemyMonController.update(enemy)
  }
  
  def onMove(move: Move): Unit = {
    implicit val rng = engine.rng
    implicit val reader = engine
    val action = Action(engine.player.lead, move, engine.player.pollTarget(move, engine.player.lead, engine.enemy))
    engine.playTurn(action)
    updateUI()
  }
  
  def !(msg: Signal): Unit = {
    msg match {
      case Message(text) => messages.text = s"${messages.text()}${text}\n"
      case DamageMsg(dmg, target, element) => playDamage(dmg, target, element)
    }
  }
  
  private def playDamage(damage: Int, target: Monster, element: Element): Unit = {
    val r = 25
    val color = Color.AliceBlue
    val center = findCenter(target)
    
    val circle = new Circle {
      radius = r
      fill = color
    }
    val points = new Group((0 until 6).map(i => {
      new Polygon {
        points ++= Seq(-r/2, r, 0, 2*r, +r/2, r)
        fill = color
        transforms += new Rotate(60 * i, 0, 0, 0, Rotate.ZAxis)
      }
    }): _*)
    val label = new Label(s"${damage}") {
      font = Font(24)
      alignment = Pos.Center
    }
    val group = new Group(circle, points, label)
    group.translateX = center.x
    group.translateY = center.y
    
    val animation = new SequentialTransition(group, Seq(
        // Fade In
        new FadeTransition{
          fromValue = 0.0
          toValue = 1.0
          duration = Duration(250)
        },
        // Fade Out
        new FadeTransition{
          fromValue = 1.0
          toValue = 0.0
          duration = Duration(750)
        }
    ))
    animation.onFinished = handle{
      animationPane.children -= group
    }
    animation.play()
    animationPane.children += group
  }
  
  private def findCenter(monster: Monster): Point2D = {
    if (monster == engine.player.lead.mon) {
      centerOfControl(playerImg)
    } else {
      centerOfControl(enemyImg)
    }
  }
  
  private def centerOfControl(control: ImageView): Point2D = {
    val bounds = control.boundsInParent()
    val x = bounds.minX + bounds.width / 2
    val y = bounds.minY + bounds.height / 2
    new Point2D(x, y)
  }
  
  private def loadMoveButton(move: Move): Button = {
    val fxmlLoader = new FXMLLoader(getClass.getResource("moveButton.fxml"))
    fxmlLoader.setBuilderFactory(new JavaFXBuilderFactory())
    val button = fxmlLoader.load[jfxsc.Button]()
    val controller = fxmlLoader.getController[MoveButton]()
    controller.setup(move, onMove)
    button
  }
}